/* Socket utility routines.
 *
 * Services is copyright (c) 1996-1998 Andy Church.
 *     E-mail: <achurch@dragonfire.net>
 * This program is free but copyrighted software; see the file COPYING for
 * details.
 */

#include "services.h"
#include <setjmp.h>


FILE **files;		/* Array of FILE *'s; files[s] = fdopen(s, "r+") */
int filescnt = 0;	/* Size of files array */

/*************************************************************************/
/*************************************************************************/

static jmp_buf alarm_jmp;

static void alarm_handler(int sig_unused)
{
    longjmp(alarm_jmp, 1);
}

/*************************************************************************/

/* Read from a socket with buffering. */

static char read_netbuf[NET_BUFSIZE];
static char *read_curpos = read_netbuf; /* Next byte to return */
static char *read_bufend = read_netbuf; /* Next position for data from socket */
static char * const read_buftop = read_netbuf + NET_BUFSIZE;

static int buffered_read(int fd, char *buf, int len)
{
    int nread, left = len;
    fd_set fds;
    struct timeval tv = {0,0};

    while (left > 0) {
	struct timeval *tvptr = (read_bufend == read_curpos ? NULL : &tv);
	FD_ZERO(&fds);
	FD_SET(fd, &fds);
	while (read_bufend != read_curpos-1 && !(read_curpos == read_netbuf && read_bufend == read_buftop-1)
				  && select(fd+1, &fds, 0, 0, tvptr) == 1) {
	    int maxread;
	    tvptr = &tv;		/* don't wait next time */
	    if (read_bufend < read_curpos)	/* wrapped around? */
		maxread = (read_curpos-1) - read_bufend;
	    else if (read_curpos == read_netbuf)
		maxread = read_buftop - read_bufend - 1;
	    else
		maxread = read_buftop - read_bufend;
	    nread = read(fd, read_bufend, maxread);
	    if (debug >= 4)
		log("debug: buffered_read wanted %d, got %d", maxread, nread);
	    if (nread <= 0)
		break;
	    read_bufend += nread;
	    if (read_bufend == read_buftop)
		read_bufend = read_netbuf;
	}
	if (read_curpos == read_bufend)	/* No more data on socket */
	    break;
	/* See if we can gobble up the rest of the buffer. */
	if (read_curpos+left >= read_buftop && read_bufend < read_curpos) {
	    nread = read_buftop-read_curpos;
	    memcpy(buf, read_curpos, nread);
	    buf += nread;
	    left -= nread;
	    read_curpos = read_netbuf;
	}
	/* Now everything we need is in a single chunk at read_curpos. */
	if (read_bufend > read_curpos && read_bufend-read_curpos < left)
	    nread = read_bufend-read_curpos;
	else
	    nread = left;
	if (nread) {
	    memcpy(buf, read_curpos, nread);
	    buf += nread;
	    left -= nread;
	    read_curpos += nread;
	}
    }
    if (debug >= 4) {
	log("debug: buffered_read(%d,%p,%d) returning %d",
			fd, buf, len, len-left);
    }
    return len - left;
}

/* Optimized version of the above for reading a single character; returns
 * the character in an int or EOF, like fgetc(). */

static int buffered_read_one(int fd)
{
    int nread;
    fd_set fds;
    struct timeval tv = {0,0};
    char c;
    struct timeval *tvptr = (read_bufend == read_curpos ? NULL : &tv);

    FD_ZERO(&fds);
    FD_SET(fd, &fds);
    while (read_bufend != read_curpos-1
		&& !(read_curpos == read_netbuf && read_bufend == read_buftop-1)
		&& select(fd+1, &fds, 0, 0, tvptr) == 1) {
	int maxread;
	tvptr = &tv;		/* don't wait next time */
	if (read_bufend < read_curpos)	/* wrapped around? */
	    maxread = (read_curpos-1) - read_bufend;
	else if (read_curpos == read_netbuf)
	    maxread = read_buftop - read_bufend - 1;
	else
	    maxread = read_buftop - read_bufend;
	nread = read(fd, read_bufend, maxread);
	if (debug >= 4)
	    log("debug: buffered_read_one wanted %d, got %d", maxread, nread);
	if (nread <= 0)
	    break;
	read_bufend += nread;
	if (read_bufend == read_buftop)
	    read_bufend = read_netbuf;
    }
    if (read_curpos == read_bufend) {	/* No more data on socket */
	if (debug >= 4)
	    log("debug: buffered_read_one(%d) returning %d", fd, EOF);
	return EOF;
    }
    c = *read_curpos++;
    if (read_curpos == read_buftop)
	read_curpos = read_netbuf;
    if (debug >= 4)
	log("debug: buffered_read_one(%d) returning %d", fd, c);
    return (int)c & 0xFF;
}

/*************************************************************************/
/*************************************************************************/

static int lastchar = EOF;

int sgetc(int s)
{
    unsigned char c;

    if (lastchar != EOF) {
	c = lastchar;
	lastchar = EOF;
	return c;
    }
    return buffered_read_one(s);
}

int sungetc(int c, int s)
{
    return lastchar = c;
}

/*************************************************************************/

/* If connection was broken, return NULL.  If the read timed out, return
 * (char *)-1. */
char *sgets(char *buf, int len, int s)
{
    int c;
    char *ptr = buf;

    if (len == 0)
	return NULL;
    if (setjmp(alarm_jmp))
	return (char *)-1;
    signal(SIGALRM, alarm_handler);
    alarm(READ_TIMEOUT);
    c = sgetc(s);
    alarm(0);
    signal(SIGALRM, SIG_IGN);
    while (--len && (*ptr++ = c) != '\n' && (c = sgetc(s)) >= 0)
	;
    if (c < 0)
	return NULL;
    *ptr = 0;
    return buf;
}

/*************************************************************************/

/* Read from a socket.  (Use this instead of read() because it has
 * buffering.) */

int sread(int s, char *buf, int len)
{
    return buffered_read(s, buf, len);
}

/*************************************************************************/

int sputs(char *str, int s)
{
    return write(s, str, strlen(str));
}

/*************************************************************************/

int sockprintf(int s, char *fmt, ...)
{
    va_list args;

    va_start(args, fmt);
    if (s >= filescnt) {
	int oldcnt = filescnt;
	filescnt *= 2;
	if (filescnt <= s)
	    filescnt = s+1;
	files = realloc(files, sizeof(FILE *) * filescnt);
	if (!files) {
	    filescnt = 0;
	    errno = ENOMEM;
	    return 0;
	}
	memset(files+oldcnt, 0, sizeof(FILE *) * (filescnt-oldcnt));
    }
    if (!files[s]) {
	if (!(files[s] = fdopen(s, "r+")))
	    return 0;
	setbuf(files[s], NULL);
    }
    return vfprintf(files[s], fmt, args);
}

/*************************************************************************/

int conn(char *host, int port)
{
    struct hostent *hp;
    struct sockaddr_in sa;
    int sock;

    if (!(hp = gethostbyname(host)))
	return -1;

    bzero(&sa, sizeof(sa));
    bcopy(hp->h_addr, (char *)&sa.sin_addr, hp->h_length);
    sa.sin_family = hp->h_addrtype;
    sa.sin_port = htons((unsigned short)port);
    if ((sock = socket(hp->h_addrtype, SOCK_STREAM, 0)) < 0)
	return -1;

    if (filescnt <= sock) {
	int oldcnt = filescnt;
	filescnt *= 2;
	if (filescnt <= sock)
	    filescnt = sock+1;
	files = realloc(files, sizeof(FILE *) * filescnt);
	if (!files) {
	    filescnt = 0;
	    shutdown(sock, 2);
	    close(sock);
	    errno = ENOMEM;
	    return -1;
	}
	memset(files+oldcnt, 0, sizeof(FILE *) * (filescnt-oldcnt));
    }
    if (!(files[sock] = fdopen(sock, "r+"))) {
	int errno_save = errno;
	shutdown(sock, 2);
	close(sock);
	errno = errno_save;
	return -1;
    }
    setbuf(files[sock], NULL);

    if (connect(sock, (struct sockaddr *)&sa, sizeof(sa)) < 0) {
	int errno_save = errno;
	shutdown(sock, 2);
	fclose(files[sock]);
	errno = errno_save;
	return -1;
    }

    return sock;
}

/*************************************************************************/

void disconn(int s)
{
    shutdown(s, 2);
    if (s < filescnt && files[s])
	fclose(files[s]);
    else
	close(s);
}

/*************************************************************************/
